import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np


class CNN(nn.Module):
    def __init__(self, device):
        super().__init__()
        self.device = device
        
        self.conv1 = nn.Conv1d(in_channels=4, out_channels=25, kernel_size=2, padding=1)
        self.pool = nn.MaxPool1d(kernel_size=2, stride=2)
        self.conv2 = nn.Conv1d(in_channels=25, out_channels=32, kernel_size=4, padding=3)
        self.fc1 = nn.Linear(992, 64)
        self.fc2 = nn.Linear(64, 1)
        self.dropout = nn.Dropout(p=0.2)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1) 
        x = F.relu(self.fc1(x))
        x = self.dropout(x)
        x = torch.sigmoid(self.fc2(x))
        return x    

class CNN_2d(nn.Module):
    def __init__(self, device):
        super().__init__()
        self.device = device
        
        self.conv1 = nn.Conv2d(in_channels=1, out_channels=25, kernel_size=(1,6), padding=1)
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)
        self.conv2 = nn.Conv2d(in_channels=25, out_channels=32, kernel_size=6, padding=3)
        self.fc1 = nn.Linear(1216, 64)
        self.fc2 = nn.Linear(64, 1)
        self.dropout = nn.Dropout(p=0.2)
        
    def forward(self, x):
        x = torch.unsqueeze(x, 1)
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x))) #when kernel size for pooling is 2

        x = torch.flatten(x, 1) 

        x = F.relu(self.fc1(x))
        x = self.dropout(x)
        x = torch.sigmoid(self.fc2(x))
        return x 


# CNN class for the hyperparameter tuning -- 1D convolutions 
class CNN_hp_tuning(nn.Module):
    def __init__(self, device, no_conv, no_filters, filter_size, stride, pool_size, pool_stride, dense_output, no_dense_layers, dropout,
                 input_length=60, input_channels = 4):
        super(CNN_hp_tuning, self).__init__()
        self.device = device
        self.no_conv = no_conv
        self.no_dense_layers = no_dense_layers        
        self.input_channels = input_channels
        self.padding = 3
        self.conv = []
        self.fc = []       
        
        for i in range(self.no_conv):
            self.conv.append(nn.Conv1d(in_channels=self.input_channels, out_channels=no_filters, kernel_size=filter_size, stride=stride,
                                       padding=self.padding))
            self.input_channels = no_filters
            #input length after one convolution and pooling layers
            input_length = ((input_length - filter_size + 2*self.padding) // stride + 1 - pool_size) // pool_stride + 1
        
        self.conv = nn.ModuleList(self.conv)
        self.pool = nn.MaxPool1d(kernel_size=pool_size, stride=pool_stride)
        dense_input = input_length*no_filters  
        
        for i in range(self.no_dense_layers):
            
            if i==self.no_dense_layers-1:
                self.fc.append(nn.Linear(dense_input, dense_output))
                dense_input = dense_output
            else:
                self.fc.append(nn.Linear(dense_input, int(dense_input/2)))
                dense_input = int(dense_input/2)
                
            
                
        self.fc = nn.ModuleList(self.fc)   
        self.fc2 = nn.Linear(dense_output, 1)
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, x):
        for i in range(self.no_conv):
            x = self.pool(F.relu(self.conv[i](x)))
            
        x = torch.flatten(x, 1) 
        
        for i in range(self.no_dense_layers):
            x = F.relu(self.fc[i](x))
            x = self.dropout(x)
        
        x = torch.sigmoid(self.fc2(x))
        return x   
    
# CNN class for the hyperparameter tuning  -- 2D convolutions  
class CNN_hp_tuning_2d(nn.Module):
    def __init__(self, device, no_conv, no_filters, filter_size, filter_size2, stride, pool_size, pool_stride, dense_output, no_dense_layers, dropout,
                 input_length=60, input_channels = 1):
        super(CNN_hp_tuning_2d, self).__init__()
        self.device = device
        self.no_conv = no_conv
        self.no_dense_layers = no_dense_layers       
        self.input_channels = input_channels        
        self.padding = (1,1)
        self.conv = []
        self.fc = []
        self.input_height = 4
        self.input_length = input_length


        #conv layer 1
        self.conv.append(nn.Conv2d(in_channels=self.input_channels, out_channels=no_filters, kernel_size=filter_size, stride=stride,
                                       padding=self.padding))
        self.input_channels = no_filters
        
        #input length after one convolution and one pooling layer
        self.input_length = ((self.input_length - filter_size[1] + 2*self.padding[1]) // stride[1] + 1 - pool_size) // pool_stride + 1
        self.input_height = ((self.input_height - filter_size[0] + 2*self.padding[0]) // stride[0] + 1 - pool_size) // pool_stride + 1
        
        self.pool = nn.MaxPool2d(kernel_size=pool_size, stride=pool_stride)   
        filter_size = (filter_size2,filter_size2)
        self.padding = (3,3)
        #conv layer 2
        self.conv.append(nn.Conv2d(in_channels=self.input_channels, out_channels=no_filters, kernel_size=filter_size, stride=stride,
                                       padding=self.padding))
        
        #input length after one convolution and one pooling layer
        self.input_length = ((self.input_length - filter_size[1] + 2*self.padding[1]) // stride[1] + 1 - pool_size) // pool_stride + 1
        self.input_height = ((self.input_height - filter_size[0] + 2*self.padding[0]) // stride[0] + 1 - pool_size) // pool_stride + 1
            
        self.conv = nn.ModuleList(self.conv)    
        
        self.dense_input = self.input_length*no_filters *self.input_height
        
        for i in range(self.no_dense_layers):            
            if i==self.no_dense_layers-1:
                self.fc.append(nn.Linear(self.dense_input, dense_output))
                self.dense_input = dense_output
            else:
                self.fc.append(nn.Linear(self.dense_input, int(self.dense_input/2)))
                self.dense_input = int(self.dense_input/2)
                
        self.fc = nn.ModuleList(self.fc)   
        self.fc2 = nn.Linear(dense_output, 1)
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, x):
       # print(self.input_length)
       # print(self.input_height)
        x = torch.unsqueeze(x, 1)
        x = self.pool(F.relu(self.conv[0](x)))
       # print(x.shape)
        x = self.pool(F.relu(self.conv[1](x)))   
       # print(x.shape)
        x = torch.flatten(x, 1) 
        
        for i in range(self.no_dense_layers):
            x = F.relu(self.fc[i](x))
            x = self.dropout(x)
        
        x = torch.sigmoid(self.fc2(x))
        return x   