import os
import numpy as np
import pandas as pd
import time
import json
import pathlib
import torch
import torch.nn as nn
import torch.optim as optim
import matplotlib.pyplot as plt
from torch.utils.data import DataLoader
from sklearn.model_selection import StratifiedShuffleSplit
from numpy import loadtxt, savetxt
from sklearn import metrics, model_selection
    
from train_utils import *
from models import *
from metrics import *

curr_dir_path = str(pathlib.Path().absolute())
cwd = os.getcwd() 
save_dir_path = '/logs/'


manual_seed = 1
if manual_seed is None:
    manual_seed = random.randint(1, 100000)
torch.manual_seed(manual_seed)
np.random.seed(manual_seed)
print("seed", manual_seed)

gpu_num = 0
cuda_available = torch.cuda.is_available()
if cuda_available:
    torch.cuda.set_device(gpu_num)


class Training():

    def __init__(self, config, save_dir = '', device='', model_no=0):
        self.config = config

        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.save_dir = save_dir
        self.model_no = model_no
        self.type_data = config['DATA']['type']
        self.model = CNN_2d(self.device)
        
        #hyperparameter tuning class test 
        #self.model = CNN_hp_tuning_2d(self.device, no_conv=config['MODEL']['Conv no'], no_filters=config['MODEL']['Filter no'], 
        #                           filter_size=config['MODEL']['Filter size'],filter_size2=config['MODEL']['Filter size2'],
        #                           stride=config['MODEL']['Conv stride'], 
        #                           pool_size=config['MODEL']['Pool size'], pool_stride=config['MODEL']['Pool stride'],
        #                           dense_output=config['MODEL']['FC layer output'], no_dense_layers = config['MODEL']['No FC layers'],
        #                           dropout=config['MODEL']['dropout'], input_length=config['DATA']['LENGTH'])
        
        print(self.model)
        
        if cuda_available:
            print("Model on GPU")
            self.model.cuda()
  
        self.optimizer = getattr(optim, config['OPTIMIZER']['type']) \
            (self.model.parameters(), lr=config['OPTIMIZER']['lr'], weight_decay=config['OPTIMIZER']['weight_decay'])

        
        self.trainloader = None
        self.metrics = {'train': {}, 'val': {}}
        self.train_accuracy_curve = []
        self.train_f1_curve = []
        self.val_accuracy_curve  = []
        self.val_f1_curve = []

    def train(self, epoch, x_train, y_train):
        
        trainset = SequenceDataset(x_train, y_train)  
        
        self.trainloader = torch.utils.data.DataLoader(
                        trainset, batch_size=self.config['DATA']['BATCH_SIZE'], shuffle=True)
                     #, num_workers=self.config['DATA']['NUM_WORKERS'])

        self.model.to(self.device)
        self.model.train()
        
        loss_function = torch.nn.BCELoss()

        # Metrics
        m = Metrics()  
        self.metrics['train'] = m.metrics      
        avg_train_loss = 0
    
        for data, labels  in self.trainloader:     
            
            self.optimizer.zero_grad()
            
            output = self.model.forward(data.to(self.device))     
            output = output.squeeze(1)
            loss = loss_function(output, labels.to(self.device))
              
            #Backpropagation + optimization step      
            loss.backward()
            self.optimizer.step()
            
            #Metrics per batch
            metrics_for_batch, pred = m.get_metrics(output.detach().clone(), labels.detach().clone())  
            for key,value in metrics_for_batch.items():
                self.metrics['train'][key] += value
                
            avg_train_loss += loss.item()
            
            
        # Metrics
        for measure in m.metrics:
            self.metrics['train'][measure] /= (len(self.trainloader)+1)
        
        avg_train_loss /= (len(self.trainloader))
        self.metrics['train']['loss'] = avg_train_loss        
        print('Epoch: {:d}, Train Loss: {:.4f}, '.format(epoch, avg_train_loss), self.metrics['train'])
        
        return self.metrics['train']

    def evaluate(self, epoch, x_val, y_val, testing=False):

        valset = SequenceDataset(x_val, y_val)  
        val_dataloader = torch.utils.data.DataLoader(
                    valset, batch_size=self.config['DATA']['BATCH_SIZE'], shuffle=True)
                  # num_workers=self.config['DATA']['NUM_WORKERS'])
        
        self.model.to(self.device)
        self.model.eval()
        
        loss_function = torch.nn.BCELoss()
        
        # Metrics
        avg_val_loss = 0
        
        m = Metrics()  
        self.metrics['val'] = m.metrics
                
        for data, labels  in val_dataloader:                 
            
            output = self.model.forward(data.to(self.device))           
            output = output.squeeze(1)
            
            loss = loss_function(output, labels.to(self.device))
            
            #Metrics per batch
            metrics_for_batch, pred = m.get_metrics(output.detach().clone(), labels.detach().clone())  
            for key,value in metrics_for_batch.items():
                self.metrics['val'][key] += value
                
            avg_val_loss += loss.item()
            
        #Metrics
        for measure in m.metrics:
            self.metrics['val'][measure] /= (len(val_dataloader)+1)
        
        avg_val_loss /= (len(val_dataloader))
        self.metrics['val']['loss'] = avg_val_loss
        
        if testing == False:
            print('Epoch: {:d}, Valid Loss: {:.4f}, '.format(epoch, avg_val_loss), self.metrics['val'])
        else:
            print('Epoch: {:d}, Test Loss: {:.4f}, '.format(epoch, avg_val_loss), self.metrics['val'])

        return self.metrics['val']


    def load_data(self):
        
        #Load data  
        print("Loading training data")
        if self.type_data == 'sliding':
            dir_data = '/mnt/sdg/angeliki/VAE_gumbel_softmax_hp_tuning-20210731T085503Z-001/VAE_gumbel_softmax_hp_tuning/pos/data/sliding/' 
            print("training data type: sliding windows")
            
            neg_seq_encoded_chr19_1 = loadtxt(dir_data + 'chr19.test_negative_spikes_uscs.csv', delimiter=',')       
            neg_seq_encoded_chr19_2 = loadtxt(dir_data + 'chr19.test_negative_spikes2_uscs.csv', delimiter=',') 
        #    pos_seq_encoded_chr19 = loadtxt(dir_data + 'chr19.test_positive_spikes.csv', delimiter=',')
        #    pos_seq_encoded_chr19_2 = loadtxt(dir_data+ 'chr19.test_positive_spikes2.csv', delimiter=',')
        #    neg_seq_encoded_chr19_3 = np.zeros((5000, neg_seq_encoded_chr19_1.shape[1]))
            pos_seq_encoded_chr19 = loadtxt(dir_data + 'chr19_sliding_positive_n.csv', delimiter=',')
            pos_seq_encoded_chr18 = loadtxt(dir_data + 'chr18_sliding_positive_n.csv', delimiter=',')
            pos_seq_encoded_chr20 = loadtxt(dir_data + 'chr20.fa_positive_sliding_80_n.csv', delimiter=',')
            
            neg_seq_encoded_total = np.concatenate((neg_seq_encoded_chr19_1, neg_seq_encoded_chr19_2), axis = 0)      
            print("neg shape", neg_seq_encoded_total.shape)
            pos_seq_encoded_total = np.concatenate((pos_seq_encoded_chr19, pos_seq_encoded_chr18, pos_seq_encoded_chr20), axis = 0)  

        elif self.type_data == '80bp':
            dir_data = '/mnt/sdg/angeliki/VAE_gumbel_softmax_hp_tuning-20210731T085503Z-001/VAE_gumbel_softmax_hp_tuning/pos/data/window_80/' 
            
            print("training data type: 80bp containing the repeat - initial position between 1-20")

            pos_seq_encoded_chr19 = loadtxt(dir_data  + 'chr19.fa_positive_80.csv', delimiter=',')  
            pos_seq_encoded_chr21 = loadtxt(dir_data +'chr21.fa_positive_80.csv', delimiter=',')
            pos_seq_encoded_chr22 = loadtxt(dir_data +'chr22.fa_positive_80.csv', delimiter=',')
            pos_seq_encoded_chr7 = loadtxt(dir_data +'chr7.fa_positive_80.csv', delimiter=',')
            pos_seq_encoded_chr16 = loadtxt(dir_data +'chr16_17_18_20.fa_positive_80.csv', delimiter=',')
            neg_seq_encoded_chr19 = loadtxt(dir_data +'chr19.fa_negative_80_uscs.csv', delimiter=',')
            neg_seq_encoded_chr19_2 = loadtxt(dir_data +'chr19.fa_negative_80_uscs2.csv', delimiter=',')
            neg_seq_encoded_chr21 = loadtxt(dir_data +'chr21.fa_negative_80_uscs1.csv', delimiter=',')
            neg_seq_encoded_chr21_2 = loadtxt(dir_data +'chr21.fa_negative_80_uscs2.csv', delimiter=',')

            #Concatenate positive and negative data 
            pos_seq_encoded_total = np.append(pos_seq_encoded_chr19, pos_seq_encoded_chr21, axis = 0)
            pos_seq_encoded_total = np.append(pos_seq_encoded_total, pos_seq_encoded_chr22, axis = 0)
            pos_seq_encoded_total = np.append(pos_seq_encoded_total, pos_seq_encoded_chr7, axis = 0)
            pos_seq_encoded_total1 = np.append(pos_seq_encoded_total, pos_seq_encoded_chr16, axis = 0)
            neg_seq_encoded_total = np.append(neg_seq_encoded_chr19, neg_seq_encoded_chr19_2, axis = 0)
            neg_seq_encoded_total = np.append(neg_seq_encoded_total, neg_seq_encoded_chr21, axis = 0)
            neg_seq_encoded_total1 = np.append(neg_seq_encoded_total, neg_seq_encoded_chr21_2, axis = 0)
            """
            dir_data = '/mnt/sdg/angeliki/VAE_gumbel_softmax_hp_tuning-20210731T085503Z-001/VAE_gumbel_softmax_hp_tuning/pos/data/sliding/' 
            print("training data type: sliding windows")
            
            neg_seq_encoded_chr19_1 = loadtxt(dir_data + 'chr19.test_negative_spikes_uscs.csv', delimiter=',')       
            neg_seq_encoded_chr19_2 = loadtxt(dir_data + 'chr19.test_negative_spikes2_uscs.csv', delimiter=',') 
            pos_seq_encoded_chr19 = loadtxt(dir_data + 'chr19.test_positive_spikes.csv', delimiter=',')
            pos_seq_encoded_chr19_2 = loadtxt(dir_data+ 'chr19.test_positive_spikes2.csv', delimiter=',')
            neg_seq_encoded_chr19_3 = np.zeros((5000, neg_seq_encoded_chr19_1.shape[1]))

            neg_seq_encoded_total = np.concatenate((neg_seq_encoded_chr19_1, neg_seq_encoded_chr19_2, neg_seq_encoded_chr19_3), axis = 0)      
            pos_seq_encoded_total = np.append(pos_seq_encoded_chr19, pos_seq_encoded_chr19_2, axis = 0)
            """
            neg_seq_encoded_total = neg_seq_encoded_total1# np.concatenate((neg_seq_encoded_total, neg_seq_encoded_total1, neg_seq_encoded_chr19_3), axis = 0)
            pos_seq_encoded_total = pos_seq_encoded_total1 #np.append(pos_seq_encoded_total, pos_seq_encoded_total1, axis = 0)
            
        N = pos_seq_encoded_total.shape[0]
        indices = np.arange(N)
        np.random.shuffle(indices)
        pos_seq_encoded_total = pos_seq_encoded_total[indices, :]
        pos_seq_encoded_total = pos_seq_encoded_total[0:500000]
        
        N = neg_seq_encoded_total.shape[0]
        indices = np.arange(N)
        np.random.shuffle(indices)
        neg_seq_encoded_total = neg_seq_encoded_total[indices, :]
        neg_seq_encoded_total = neg_seq_encoded_total[0:500000]

        print('positive size:', pos_seq_encoded_total.shape)
        print('negative size:', neg_seq_encoded_total.shape)
        
        seq_encoded = np.append(pos_seq_encoded_total, neg_seq_encoded_total, axis = 0)

        y_label_pos = np.ones(pos_seq_encoded_total.shape[0])  #positive labels
        y_label_neg = np.zeros(neg_seq_encoded_total.shape[0])  #negative labels
        y_label = np.append(y_label_pos, y_label_neg)
        print(seq_encoded.shape, y_label.shape)       
        
        #Shuffle the data (positive and negative examples should be shuffled)
        N = y_label.shape[0]
        indices = np.arange(N)
        np.random.shuffle(indices)
        seq_encoded = seq_encoded[indices, :]
        y_label = y_label[indices]
        
        # Split (train+val) and test  
        # Stratified splits 
        train_val_data, test_data, train_val_labels, test_labels = model_selection.train_test_split(seq_encoded, y_label, test_size=0.05,
                                                                                                    random_state=1, stratify=y_label)
        if self.config['VALIDATION']['apply']:
            # Split train and validation set 
            train_data, val_data, train_labels, val_labels = model_selection.train_test_split(train_val_data,
                                                                                              train_val_labels,test_size=0.02,
                                                                                              random_state=1, stratify=train_val_labels)
        else:
            train_data, train_labels = train_val_data, train_val_labels

        # Reshape training image
        length = int(train_data.shape[1]/4)
        train_data = train_data.reshape(-1, length, 4)
        val_data = val_data.reshape(-1, length, 4)
        test_data = test_data.reshape(-1, length, 4)
        
        #Shuffle the repeats
        if self.config['DATA']['SHUFFLE_REPEATS']:
            print("shuffling the repeats")
            train_data = random_shuffle(train_data) 
            val_data = random_shuffle(val_data) 
            test_data = random_shuffle(test_data) 
            
        train_data = np.transpose(train_data, (0, 2, 1))
        val_data = np.transpose(val_data, (0, 2, 1))
        test_data = np.transpose(test_data, (0, 2, 1))
       
        print('Training data shape : ', train_data.shape, train_labels.shape)
        print('Validation data shape : ', val_data.shape, val_labels.shape)
        print('Testing data shape : ', test_data.shape, test_labels.shape)
        
        return train_data, train_labels, val_data, val_labels, test_data, test_labels
    
    def save_data(self):
        # Save metrics and model
        if self.config['TRAINER']["save_model"]:
            json_dir = cwd + self.save_dir + f"model{self.model_no}.pt"
            print("Saving data at: " + json_dir)
            torch.save(self.model.state_dict(), json_dir)
            #to load model:
            #json_dir = self.save_dir + "model.pt"
            #model = CNN(args)
            #model.load_state_dict(torch.load(json_dir))
            
            #Save final metrics 
            results = [self.metrics, self.config]
            json_dir = cwd + self.save_dir + "final_metrics"
            print("Saving data at: " + json_dir)
            output_folder = pathlib.Path(json_dir)
            output_folder.mkdir(parents=True, exist_ok=True)
            results_json = json.dumps(results, indent=4, sort_keys=True)
            (output_folder / f"results{self.model_no}.json").write_text(results_json)
            
            train_accuracy_curve = np.hstack(self.train_accuracy_curve)
            train_f1_curve = np.hstack(self.train_f1_curve)
            val_accuracy_curve = np.hstack(self.val_accuracy_curve)
            val_f1_curve = np.hstack(self.val_f1_curve)
            #Save training and validation curves (accuracy, f1)
            results = {"train_accuracy": train_accuracy_curve.tolist(), "train_f1":  train_f1_curve.tolist(), 
                        "val_accuracy": val_accuracy_curve.tolist(), "val_f1":  val_f1_curve.tolist()}            
            json_dir_curve = cwd + self.save_dir + "curves"
            print("Saving training and validation curves at: " + json_dir_curve)
            output_folder = pathlib.Path(json_dir_curve)
            output_folder.mkdir(parents=True, exist_ok=True)
            results_json = json.dumps(results, indent=4, sort_keys=True)
            (output_folder / f"results{self.model_no}.json").write_text(results_json)
            
            
    def training_pipeline(self):
        
        train_accuracy_curve = []
        train_f1_curve = []
        val_accuracy_curve = []
        val_f1_curve = []
        x_train, y_train, x_val, y_val, x_test, y_test = self.load_data()
           
        for epoch in range(self.config['TRAINER']['epochs']):
            print("Training Epoch %i -------------------" % epoch)

            epoch_tic = time.time()
            metrics = self.train(epoch, x_train, y_train)
            self.train_accuracy_curve.append(metrics['acc'])
            self.train_f1_curve.append(metrics['f1'])
            
            if self.config['VALIDATION']['apply']:
                metrics = self.evaluate(epoch, x_val, y_val)
                self.val_accuracy_curve.append(metrics['acc'])
                self.val_f1_curve.append(metrics['f1'])
                
            epoch_toc = time.time()
            epoch_time = epoch_toc - epoch_tic

            print("--------------Epoch %i completed in %i seconds--------------" % (epoch, epoch_time))
          
        #Testing
        metrics = self.evaluate(epoch, x_test, y_test, testing = True)
        self.save_data()
        
        return self.train_accuracy_curve, self.train_f1_curve, self.val_accuracy_curve, self.val_f1_curve
    

if __name__ == "__main__":

    chrm =  "21_22/"

    # Get config file
    with open(curr_dir_path + "/config.json", encoding='utf-8', errors='ignore') as json_data:
        config = json.load(json_data, strict=False)   

    config['TRAINER']['save_dir'] = save_dir_path
    obj = Training(config, save_dir_path)
    train_accuracy_curve, train_f1_curve, val_accuracy_curve, val_f1_curve = obj.training_pipeline()
     
    #Plot training and validation curves (accuracy, f1)
    plt.plot(train_accuracy_curve, label='train_accuracy')
    plt.plot(val_accuracy_curve, label = 'val_accuracy')
    plt.plot(train_f1_curve, label='train_f1')
    plt.plot(val_f1_curve, label = 'val_f1')
    plt.xlabel('Epoch')
    plt.ylabel('Accuracy')
    plt.ylim([0.5, 1.0])
    plt.legend(loc='lower right') 
       