import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np

class MLP(nn.Module):
    def __init__(
          self,
          input_dim=16,
          features=30,
          depth=0,
          args = None,
          num_outputs=1
    ):
        super().__init__()
        
        
        self.num_outputs = num_outputs
        self.input_dim = input_dim
        self.first = nn.Linear(input_dim, features)
        self.layers = nn.ModuleList(
            [nn.Linear(features, features) for i in range(depth)]
        )
        self.sigm = torch.nn.Sigmoid()
        if num_outputs is not None:
            self.last = nn.Linear(features, num_outputs)
        
    def forward(self, x):
        x = x.view(-1, self.input_dim)
        x = self.first(x)
        x = F.relu(x)
        
        for layer in self.layers:
            
            x = layer(x)
            x = F.relu(x)

        if self.num_outputs is not None:
            x = self.last(x)
        x = self.sigm(x)
        
        return x
    
class CNN(nn.Module):
    def __init__(self, device):
        super().__init__()
        self.device = device
        
        self.conv1 = nn.Conv1d(in_channels=4, out_channels=25, kernel_size=2, padding=1)
        self.pool = nn.MaxPool1d(kernel_size=2, stride=2)
        self.conv2 = nn.Conv1d(in_channels=25, out_channels=32, kernel_size=4, padding=3)
        self.fc1 = nn.Linear(992, 64)
        self.fc2 = nn.Linear(64, 1)
        self.dropout = nn.Dropout(p=0.2)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1) 
        x = F.relu(self.fc1(x))
        x = self.dropout(x)
        x = torch.sigmoid(self.fc2(x))
        return x    
    
# CNN class for the hyperparameter tuning    
class CNN_hp_tuning(nn.Module):
    def __init__(self, device, no_conv, no_filters, filter_size, stride, pool_size, pool_stride, dense_output, no_dense_layers, dropout,
                 input_length=120, input_channels = 4):
        super().__init__()
        self.device = device
        self.no_conv = no_conv
        self.no_dense_layers = no_dense_layers
        self.conv = []
        self.fc = []
        
        self.input_channels = input_channels
        self.padding = 3
       
        
        for i in range(self.no_conv):
            self.conv.append(nn.Conv1d(in_channels=self.input_channels, out_channels=no_filters, kernel_size=filter_size, stride=stride,
                                       padding=self.padding))
            self.input_channels = no_filters
            #input length after one convolution and pooling layers
            input_length = ((input_length - filter_size + 2*self.padding) // stride + 1 - pool_size) // pool_stride + 1
        
        self.conv = nn.ModuleList(self.conv)
        self.pool = nn.MaxPool1d(kernel_size=pool_size, stride=pool_stride)
        dense_input = input_length*no_filters  
        
        for i in range(self.no_dense_layers):
            self.fc.append(nn.Linear(dense_input, dense_output))
            dense_input = dense_output
         #   if (i == no_dense_layers - 1):
         #       dense_output = 1
                
        self.fc = nn.ModuleList(self.fc)   
        self.fc2 = nn.Linear(dense_output, 1)
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, x):
        for i in range(self.no_conv):
            x = self.pool(F.relu(self.conv[i](x)))
            
        x = torch.flatten(x, 1) 
        
        for i in range(self.no_dense_layers):
            x = F.relu(self.fc[i](x))
        
        x = self.dropout(x)
        x = torch.sigmoid(self.fc2(x))
        return x   